from __future__ import annotations

import urwid

from todoman import widgets

_palette = [("error", "light red", "")]


class TodoEditor:
    """
    The UI for a single todo entry.
    """

    def __init__(self, todo, lists, formatter):
        """
        :param model.Todo todo: The todo object which will be edited.
        """
        self.current_list = todo.list
        self.todo = todo
        self.lists = list(lists)
        self.formatter = formatter
        self._loop = None

        self._status = urwid.Text("")
        self._init_basic_fields()
        self._init_list_selector()
        self._init_help_text()

        save_btn = urwid.Button("Save", on_press=self._save)
        cancel_text = urwid.Text("Hit Ctrl-C to cancel, F1 for help.")
        buttons = urwid.Columns([(8, save_btn), cancel_text], dividechars=2)

        pile_items = []
        for label, field in [
            ("Summary", self._summary),
            ("Description", self._description),
            ("Location", self._location),
            ("Categories", self._categories),
            ("Start", self._dtstart),
            ("Due", self._due),
            ("Completed", self._completed),
            ("Priority", self._priority),
        ]:
            label = urwid.Text(label + ":", align="right")
            column = urwid.Columns([(13, label), field], dividechars=1)
            pile_items.append(("pack", column))

        grid = urwid.Pile(pile_items)
        spacer = urwid.Divider()

        self.left_column = urwid.ListBox(
            urwid.SimpleListWalker([grid, spacer, self._status, buttons])
        )
        right_column = urwid.ListBox(
            urwid.SimpleListWalker([urwid.Text("List:\n"), *self.list_selector])
        )

        self._ui = urwid.Columns([self.left_column, right_column])

    def _init_basic_fields(self):
        self._summary = widgets.ExtendedEdit(
            parent=self,
            edit_text=self.todo.summary,
        )
        self._description = widgets.ExtendedEdit(
            parent=self,
            edit_text=self.todo.description,
            multiline=True,
        )
        self._categories = widgets.ExtendedEdit(
            parent=self,
            edit_text=self.formatter.format_categories(self.todo.categories),
        )
        self._location = widgets.ExtendedEdit(
            parent=self,
            edit_text=self.todo.location,
        )
        self._due = widgets.ExtendedEdit(
            parent=self,
            edit_text=self.formatter.format_datetime(self.todo.due),
        )
        self._dtstart = widgets.ExtendedEdit(
            parent=self,
            edit_text=self.formatter.format_datetime(self.todo.start),
        )
        self._completed = urwid.CheckBox("", state=self.todo.is_completed)
        self._priority = widgets.PrioritySelector(
            parent=self,
            priority=self.todo.priority,
            formatter_function=self.formatter.format_priority,
        )

    def _init_list_selector(self):
        self.list_selector = []
        for _list in self.lists:
            urwid.RadioButton(
                self.list_selector,
                _list.name,
                state=_list == self.current_list,
                on_state_change=self._change_current_list,
                user_data=_list,
            )

    def _init_help_text(self):
        self._help_text = urwid.Text(
            "\n\n"
            "Global:\n"
            " F1: Toggle help\n"
            " Ctrl-C: Cancel\n"
            " Ctrl-S: Save (only works if not a shell shortcut already)\n"
            "\n"
            "In Textfields:\n"
            + "\n".join(f" {k}: {v}" for k, v in widgets.ExtendedEdit.HELP)
            + "\n\nIn Priority Selector:\n"
            + "\n".join(f" {k}: {v}" for k, v in widgets.PrioritySelector.HELP)
        )

    def _change_current_list(self, radio_button, new_state, new_list):
        if new_state:
            self.current_list = new_list

    def _toggle_help(self):
        if self.left_column.body.contents[-1] is self._help_text:
            self.left_column.body.contents.pop()
        else:
            self.left_column.body.contents.append(self._help_text)
        self._loop.draw_screen()

    def set_status(self, text):
        self._status.set_text(text)

    def edit(self):
        """Shows the UI for editing a given todo."""
        self._loop = urwid.MainLoop(
            self._ui,
            palette=_palette,
            unhandled_input=self._keypress,
            handle_mouse=False,
        )
        try:
            self._loop.run()
        except KeyboardInterrupt:
            self._loop.stop()  # Try to leave terminal in usable state
        self._loop = None

    def _save(self, btn=None):
        try:
            self._save_inner()
        except Exception as e:
            self.set_status(("error", str(e)))
        else:
            raise urwid.ExitMainLoop

    def _save_inner(self):
        self.todo.list = self.current_list
        self.todo.summary = self.summary
        self.todo.description = self.description
        self.todo.location = self.location
        self.todo.due = self.formatter.parse_datetime(self.due)
        self.todo.start = self.formatter.parse_datetime(self.dtstart)
        if not self.todo.is_completed and self._completed.get_state():
            self.todo.complete()
        elif self.todo.is_completed and not self._completed.get_state():
            self.todo.status = "NEEDS-ACTION"
            self.todo.completed_at = None
        self.todo.categories = [c.strip() for c in self.categories.split(",")]
        self.todo.priority = self.priority

        # TODO: categories
        # TODO: comment

        # https://tools.ietf.org/html/rfc5545#section-3.8
        # geo (lat, lon)
        # RESOURCE: the main room

    def _keypress(self, key):
        if key.lower() == "f1":
            self._toggle_help()
        elif key == "ctrl s":
            self._save()

    @property
    def summary(self):
        return self._summary.edit_text

    @property
    def description(self):
        return self._description.edit_text

    @property
    def location(self):
        return self._location.edit_text

    @property
    def due(self):
        return self._due.edit_text

    @property
    def dtstart(self):
        return self._dtstart.edit_text

    @property
    def categories(self):
        return self._categories.edit_text

    @property
    def priority(self):
        return self._priority.priority



from __future__ import annotations


class TodomanError(Exception):
    """
    Base class for all our exceptions.

    Should not be raised directly.
    """


class NoSuchTodoError(TodomanError):
    EXIT_CODE = 20

    def __str__(self):
        return f"No todo with id {self.args[0]}."


class ReadOnlyTodoError(TodomanError):
    EXIT_CODE = 21

    def __str__(self):
        return (
            "Todo is in read-only mode because there are multiple todosin {}.".format(
                self.args[0]
            )
        )


class NoListsFoundError(TodomanError):
    EXIT_CODE = 22

    def __str__(self):
        return "No lists found matching {}, create a directory for a new list.".format(
            self.args[0]
        )


class AlreadyExistsError(TodomanError):
    """
    Raised when two objects have a same identity.

    This can ocurrs when two lists have the same name, or when two Todos have
    the same path.
    """

    EXIT_CODE = 23

    def __str__(self):
        return "More than one {} has the same identity: {}.".format(*self.args)



from __future__ import annotations

import os
from importlib.util import module_from_spec
from importlib.util import spec_from_file_location
from os.path import exists
from os.path import join
from typing import Any
from typing import Callable
from typing import NamedTuple

import xdg.BaseDirectory

from todoman import __documentation__


def expand_path(path: str) -> str:
    """expands `~` as well as variable names"""
    return os.path.expanduser(os.path.expandvars(path))


def validate_cache_path(path: str) -> str:
    path = path.replace("$XDG_CACHE_HOME", xdg.BaseDirectory.xdg_cache_home)
    return expand_path(path)


def validate_date_format(fmt: str) -> str:
    if any(x in fmt for x in ("%H", "%M", "%S", "%X")):
        raise ConfigurationError(
            "Found time component in `date_format`, please use `time_format` for that."
        )
    return fmt


def validate_time_format(fmt: str) -> str:
    if any(x in fmt for x in ("%Y", "%y", "%m", "%d", "%x")):
        raise ConfigurationError(
            "Found date component in `time_format`, please use `date_format` for that."
        )
    return fmt


def validate_color_config(value: str) -> str:
    if value not in ["always", "auto", "never"]:
        raise ConfigurationError("Invalid `color` settings.")
    return value


def validate_default_priority(value: int | None) -> int | None:
    if value and not (0 <= value <= 9):
        raise ConfigurationError("Invalid `default_priority` settings.")
    return value


class ConfigEntry(NamedTuple):
    name: str
    type: type | tuple[type]
    default: Any
    description: str
    validation: Callable | None


NO_DEFAULT = object()

# A list of tuples (name, type, default, description, validation)
CONFIG_SPEC: list[ConfigEntry] = [
    ConfigEntry(
        "path",
        str,
        NO_DEFAULT,
        """
A glob pattern matching the directories where your todos are located. This
pattern will be expanded, and each matching directory (with any icalendar
files) will be treated as a list.""",
        expand_path,
    ),
    ConfigEntry(
        "color",
        str,
        "auto",
        """
By default todoman will disable colored output if stdout is not a TTY (value
``auto``). Set to ``never`` to disable colored output entirely, or ``always``
to enable it regardless. This can be overridden with the ``--color`` option.
""",
        validate_color_config,
    ),
    ConfigEntry(
        "date_format",
        str,
        "%x",
        """
The date format used both for displaying dates, and parsing input dates. If
this option is not specified the system locale's is used.
""",
        validate_date_format,
    ),
    ConfigEntry(
        "time_format",
        str,
        "%X",
        """
The date format used both for displaying times, and parsing input times.
""",
        validate_time_format,
    ),
    ConfigEntry(
        "dt_separator",
        str,
        " ",
        """
The string used to separate date and time when displaying and parsing.
""",
        None,
    ),
    ConfigEntry(
        "humanize",
        bool,
        False,
        """
If set to true, datetimes will be printed in human friendly formats like
"tomorrow", "in one hour", "3 weeks ago", etc.

If false, datetimes will be formatted using ``date_format`` and
``time_format``.
""",
        None,
    ),
    ConfigEntry(
        "default_list",
        (str, None.__class__),  # type: ignore[arg-type]
        None,
        """
The default list for adding a todo. If you do not specify this option, you
must use the ``--list`` / ``-l`` option every time you add a todo.
""",
        None,
    ),
    ConfigEntry(
        "default_due",
        int,
        24,
        """
The default difference (in hours) between new todo's due date and creation
date. If not specified, the value is 24. If set to 0, the due date for new
todos will not be set.
""",
        None,
    ),
    ConfigEntry(
        "cache_path",
        str,
        "$XDG_CACHE_HOME/todoman/cache.sqlite3",
        """
The location of the cache file (an sqlite database). This file is used to
store todo data and speed up execution/startup, and also contains the IDs for
todos.
If the value is not specified, a path relative to ``$XDG_CACHE_HOME`` will be used.
``$XDG_CACHE_HOME`` generally resolves to ``~/.cache/``.
""",
        validate_cache_path,
    ),
    ConfigEntry(
        "startable",
        bool,
        False,
        """
If set to true, only show todos which are currently startable; these are
todos which have a start date today, or some day in the past.  Todos with no
start date are always considered current. Incomplete todos (eg:
partially-complete) are also included.
""",
        None,
    ),
    ConfigEntry(
        "default_command",
        str,
        "list",
        """
When running ``todo`` with no commands, run this command.
""",
        None,
    ),
    ConfigEntry(
        "default_priority",
        (int, None.__class__),  # type: ignore[arg-type]
        None,
        """
The default priority of a task on creation.
Highest priority is 1, lowest priority is 10, and 0 means no priority at all.
""",
        validate_default_priority,
    ),
]


class ConfigurationError(Exception):
    def __init__(self, msg: str) -> None:
        super().__init__(
            f"{msg}\nFor details on the configuration format and a sample file, "
            f"see\n{__documentation__}configure.html"
        )


def find_config(config_path: str | None = None) -> str:
    if not config_path:
        for d in xdg.BaseDirectory.xdg_config_dirs:
            path = join(d, "todoman", "config.py")
            if exists(path):
                config_path = path
                break

    if not config_path:
        raise ConfigurationError("No configuration file found.\n\n")
    if not exists(config_path):
        raise ConfigurationError(f"Configuration file {config_path} does not exist.\n")
    return config_path


def load_config(custom_path: str | None = None) -> dict:
    path = find_config(custom_path)
    spec = spec_from_file_location("config", path)
    if not spec or not spec.loader:
        raise ConfigurationError("Failed to parse file as spec")
    config_source = module_from_spec(spec)

    spec.loader.exec_module(config_source)

    # TODO: Handle SyntaxError

    config = {}
    for name, type_, default, _description, validation in CONFIG_SPEC:
        value = getattr(config_source, name, default)
        if value == NO_DEFAULT:
            raise ConfigurationError(f"Missing '{name}' setting.")
        if not isinstance(value, type_):
            if isinstance(type_, tuple):
                expected = ",".join([t.__name__ for t in type_])
            else:
                expected = type_.__name__
            actual = value.__class__.__name__
            raise ConfigurationError(
                f"Bad {name} setting. Invalid type "
                f"(expected {expected}, got {actual})."
            )
        if validation:
            value = validation(value)
        config[name] = value

    return config



# Copyright (c) 2016-2023 Hugo Osvaldo Barrera
# Copyright (c) 2013-2016 Christian Geier et al.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import re
from typing import Callable

import click
import urwid


class ExtendedEdit(urwid.Edit):
    """A text editing widget supporting some more editing commands"""

    HELP = (
        ("Ctrl-W", "Delete word"),
        ("Ctrl-U", "Delete until beginning of line"),
        ("Ctrl-K", "Delete until end of line"),
        ("Ctrl-A", "Go to beginning of line"),
        ("Ctrl-E", "Go to end of line"),
        ("Ctrl-D", "Delete forward letter"),
        ("Ctrl-O", "Edit in $EDITOR"),
    )

    def __init__(self, parent: urwid.Widget, *a, **kw) -> None:
        self._parent = parent
        super().__init__(*a, **kw)

    def keypress(
        self,
        size: tuple[int, int] | tuple[int] | tuple[()],
        key: str,
    ) -> None:
        if key == "ctrl w":
            self._delete_word()
        elif key == "ctrl u":
            self._delete_till_beginning_of_line()
        elif key == "ctrl k":
            self._delete_till_end_of_line()
        elif key == "ctrl a":
            self._goto_beginning_of_line()
        elif key == "ctrl e":
            self._goto_end_of_line()
        elif key == "ctrl d":
            self._delete_forward_letter()
        elif key == "ctrl o":
            # Allow editing in $EDITOR
            self._editor()
        # TODO: alt b, alt f
        else:
            return super().keypress(size, key)

        return None

    def _delete_forward_letter(self) -> None:
        text = self.get_edit_text()
        pos = self.edit_pos
        text = text[:pos] + text[pos + 1 :]
        self.set_edit_text(text)

    def _delete_word(self) -> None:
        """delete word before cursor"""
        text = self.get_edit_text()
        t = text[: self.edit_pos].rstrip()

        words = re.findall(r"[\w]+|[^\w\s]", t, re.UNICODE)
        f_text = t if t == "" else t[: len(t) - len(words[-1])]

        self.set_edit_text(f_text + text[self.edit_pos :])
        self.set_edit_pos(len(f_text))

    def _delete_till_beginning_of_line(self) -> None:
        """delete till start of line before cursor"""
        text = self.get_edit_text()
        sol = text.rfind("\n", 0, self.edit_pos) + 1

        before_line = text[:sol]

        self.set_edit_text(before_line + text[self.edit_pos :])
        self.set_edit_pos(sol)

    def _delete_till_end_of_line(self) -> None:
        """delete till end of line before cursor"""
        text = self.get_edit_text()
        eol = text.find("\n", self.edit_pos)

        after_eol = "" if eol == -1 else text[eol:]

        self.set_edit_text(text[: self.edit_pos] + after_eol)

    def _goto_beginning_of_line(self) -> None:
        text = self.get_edit_text()
        sol = text.rfind("\n", 0, self.edit_pos) + 1
        self.set_edit_pos(sol)

    def _goto_end_of_line(self) -> None:
        text = self.get_edit_text()
        eol = text.find("\n", self.edit_pos)
        if eol == -1:
            eol = len(text)
        self.set_edit_pos(eol)

    def _editor(self) -> None:
        self._parent._loop.screen.clear()
        new_text = click.edit(self.get_edit_text())
        if new_text is not None:
            self.set_edit_text(new_text.strip())


class PrioritySelector(urwid.Button):
    HELP = (
        ("left", "Lower Priority"),
        ("right", "Higher Priority"),
    )

    RANGES = (
        [0],
        [9, 8, 7, 6],
        [5],
        [1, 2, 3, 4],
    )

    def __init__(
        self,
        parent: urwid.Widget,
        priority: int | None,
        formatter_function: Callable[[int | None], str],
    ) -> None:
        self._parent = parent
        self._label = urwid.SelectableIcon("", 0)
        urwid.WidgetWrap.__init__(self, self._label)

        self._priority = priority
        self._formatter = formatter_function
        self._set_label()

    def _set_label(self) -> None:
        self.set_label(self._formatter(self._priority))

    def _update_label(self, delta: int = 0) -> None:
        for i, r in enumerate(PrioritySelector.RANGES):
            if self._priority in r:
                self._priority = PrioritySelector.RANGES[
                    (i + delta) % len(PrioritySelector.RANGES)
                ][0]
                self._set_label()
                return

    def keypress(
        self,
        size: tuple[int, int] | tuple[int] | tuple[()],
        key: str,
    ) -> None:
        if key in ["right", "enter"]:
            self._update_label(1)
            return None
        if key == "left":
            self._update_label(-1)
            return None

        return super().keypress(size, key)

    @property
    def priority(self) -> int | None:
        return self._priority
